# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

# ruff: noqa: TID251

from __future__ import annotations

import dataclasses
import functools
import inspect
import logging
import types
import typing
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
)

import onnx
import onnx.defs
import onnx_ir as ir
from typing_extensions import ParamSpec

from onnxscript._internal import ast_utils, irbuilder, sourceinfo
from onnxscript._internal import converter as converter_module
from onnxscript.ir import _schemas
from onnxscript.onnx_types import ONNXType

_R = TypeVar("_R")
_P = ParamSpec("_P")


def select_ir_version(version: int, domain: str = "") -> int:
    """Selects a suitable ONNX ir_version for a given opset version."""
    if domain == "":
        domain = "ai.onnx"
    if (domain, version) not in onnx.helper.OP_SET_ID_VERSION_MAP:
        return max(
            v for k, v in onnx.helper.OP_SET_ID_VERSION_MAP.items() if k[0] == "ai.onnx"
        )
    required_min_version = onnx.helper.OP_SET_ID_VERSION_MAP[domain, version]
    return max(required_min_version, 10)


_ATTRIBUTE_TYPE_TO_PYTHON_TYPE = {
    onnx.defs.OpSchema.AttrType.FLOAT: float,
    onnx.defs.OpSchema.AttrType.INT: int,
    onnx.defs.OpSchema.AttrType.STRING: str,
    onnx.defs.OpSchema.AttrType.TENSOR: None,
    onnx.defs.OpSchema.AttrType.GRAPH: None,
    onnx.defs.OpSchema.AttrType.SPARSE_TENSOR: None,
    onnx.defs.OpSchema.AttrType.TYPE_PROTO: None,
    onnx.defs.OpSchema.AttrType.FLOATS: Sequence[float],
    onnx.defs.OpSchema.AttrType.INTS: Sequence[int],
    onnx.defs.OpSchema.AttrType.STRINGS: Sequence[str],
    onnx.defs.OpSchema.AttrType.TENSORS: None,
    onnx.defs.OpSchema.AttrType.GRAPHS: None,
    onnx.defs.OpSchema.AttrType.SPARSE_TENSORS: None,
    onnx.defs.OpSchema.AttrType.TYPE_PROTOS: None,
}

# A special value to indicate that the default value is not specified
_EmptyDefault = object()

logger = logging.getLogger(__name__)


class Opset:
    """Represents an ONNX Opset, which consists of a domain name, a version.

    It also contains a set of operations. This represents an Opset defined
    in the ONNX schema registry and the operations are retrieved from the
    ONNX schema registry. It also stores function definitions created for
    ops in the corresponding Opset.

    Only a single instance of Opset is created for a given (domain, version) pair.
    """

    domain: str
    version: int
    cache: ClassVar[dict[tuple[type, str, int], Opset]] = {}

    def __new__(cls, domain: str, version: int):
        key = (cls, domain, version)
        existing = cls.cache.get(key)
        if existing:
            return existing
        instance = super().__new__(cls)
        instance.domain = domain  # type: ignore[attr-defined]
        instance.version = version  # type: ignore[attr-defined]
        instance.function_defs = {}  # type: ignore[attr-defined]
        cls.cache[key] = instance
        return instance

    def __init__(self, domain: Optional[str] = None, version: Optional[int] = None):
        # Nothing to do. Object is initialized by __new__
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}({self.domain!r}, {self.version!r})"

    def __getitem__(self, opname):
        try:
            schema = onnx.defs.get_schema(opname, self.version, self.domain)
            return Op(self, opname, schema)
        except Exception:  # pylint: disable=broad-except # TODO: more specific exception
            return None

    def __contains__(self, opname):
        try:
            onnx.defs.get_schema(opname, self.version, self.domain)
        except Exception:  # pylint: disable=broad-except # TODO: more specific exception
            return False
        else:
            return True

    def __str__(self) -> str:
        return self.domain

    def __getattr__(self, attr: str) -> Op:
        try:
            schema = onnx.defs.get_schema(attr, self.version, self.domain)
            return Op(self, attr, schema)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise AttributeError(f"Attribute {attr} not found.") from exc

    def add_function_def(self, fun):
        self.function_defs[fun.name] = fun

    def _prepare_inputs(self, _: onnx.defs.OpSchema, *inputs):
        """Trims 'None' values from the end of the inputs list. This is used to support
        omitting optional inputs when no more required inputs follow to prepare a valid call
        against the Op. Used by the static opset code generator.
        """
        # TODO: validate the op schema as 'None' values are removed?
        input_list = list(inputs)
        while input_list and input_list[-1] is None:
            input_list.pop()
        return input_list


# ONNX ops


def _get_attribute_value(attr_proto: onnx.AttributeProto) -> Any:
    """Get the default value of an ONNX attribute."""
    if attr_proto.type == onnx.AttributeProto.UNDEFINED:
        return _EmptyDefault
    return onnx.helper.get_attribute_value(attr_proto)


def _typeinfo(var: ir.Value) -> Any:
    return var.meta.get("typeinfo")


@typing.runtime_checkable
class OpLike(Protocol):
    """A protocol for objects that have an ONNX OpSchema."""

    @property
    def name(self) -> str: ...

    @property
    def opset(self) -> Opset: ...

    @property
    def op_signature(self) -> Optional[ir.schemas.OpSignature]: ...


class Op(OpLike):
    """Represents an ONNX op instance (for example, the MatMul op from ONNX opset version 13).

    It belongs to a particular Opset and has a name.

    Attributes:
        opset: The Opset that this op belongs to.
        name: The name of the op.
        op_schema: The ONNX OpSchema for the op.
    """

    def __init__(
        self, opset: Opset, name: str, op_schema: Optional[onnx.defs.OpSchema] = None
    ) -> None:
        self._opset = opset
        self._name = name
        self._op_schema: onnx.defs.OpSchema | None
        if op_schema is not None:
            self._op_schema = op_schema
        elif (op := opset[name]) is not None:
            self._op_schema = op.op_schema
        else:
            self._op_schema = None
        self._signature: Optional[ir.schemas.OpSignature] = None

        if self._op_schema is None:
            logger.debug(
                "An OpSchema was not provided for Op '%s' and "
                "there is not one found in opset '%s'.",
                name,
                opset,
            )

    def __call__(self, *args, **kwargs):
        from onnxscript._internal import evaluator  # pylint: disable=import-outside-toplevel

        default_evaluator = evaluator.default()
        if hasattr(default_evaluator, "eval"):
            # Interface prior to onnxscript 0.6, used by PyTorch 2.10 and older
            if self.op_schema is None:
                raise ValueError(f"OpSchema not found for op '{self.name}'.")
            return default_evaluator.eval(self.op_schema, args, kwargs)
        # Use the new interface
        return evaluator.default().eval_op(self, args, kwargs)

    @property
    def name(self) -> str:
        return self._name

    @property
    def domain(self) -> str:
        return self._opset.domain

    @property
    def opset(self) -> Opset:
        return self._opset

    @property
    def op_schema(self) -> Optional[onnx.defs.OpSchema]:
        return self._op_schema

    @property
    def op_signature(self) -> Optional[ir.schemas.OpSignature]:
        """Returns the signature of this op."""
        if self._signature is not None:
            return self._signature

        if self.op_schema is None:
            return None

        self._signature = ir.schemas.OpSignature.from_op_schema(self.op_schema)
        return self._signature

    @op_signature.setter
    def op_signature(self, value: ir.schemas.OpSignature):
        self._signature = value


@dataclasses.dataclass(repr=False, eq=False)
class OnnxClosure:
    """Represents a nested function used as a graph-valued attribute for an ONNX op call."""

    function_ir: irbuilder.IRFunction

    # frame is python's stack-frame for the execution of top-level
    # script function (in eager-mode). It is used to get the current
    # value of outer-scope variables referred to inside this nested
    # function/GraphProto.
    frame: types.FrameType

    function: Any


class OnnxFunction(Op, Generic[_P, _R]):
    """Represents an ONNX op for which a function-body has been defined in onnxscript.

    Attributes:
        opset: Opset the function belongs to.
        name: Name of the function.
        function: Python function.
        function_ir: Python code parsed as an :class:`irbuilder.IRFunction`.
        source: Source code used to generate the function.
        kwargs: Additional properties used to construct a ModelProto.
        op_schema: Generated ONNX OpSchema for this op.
    """

    def __init__(
        self,
        opset: Optional[Opset],
        pyfun: Callable,
        irfun: irbuilder.IRFunction,
        source: str,
        kwargs: dict[str, Any],
    ):
        """Constructs an OnnxFunction.

        Args:
            opset: opset the function belongs to
            pyfun: python function
            irfun: python code parsed by class
                :class:`onnxscript.converter.Converter`
            source: source code used to generate the function
            kwargs: additional properties used to construct a ModelProto
        """
        opset = opset or Opset(irfun.domain, 1)
        super().__init__(opset, irfun.name)
        self.function = pyfun
        self.function_ir = irfun
        # Record the function domain's opset version in the function metadata
        self.function_ir.meta["opset_version"] = opset.version
        self.source = source
        self.kwargs = kwargs
        self._signature = _schemas.op_signature_from_function(
            self.function, domain=self.function_ir.domain, name=self.name
        )

        # Allow the object to be inspected as a function
        functools.update_wrapper(self, pyfun)

        # Experimental fields
        self.traceable = False

    @property
    def op_signature(self) -> Optional[ir.schemas.OpSignature]:
        """Returns the signature of this op."""
        return self._signature

    @op_signature.setter
    def op_signature(self, value: ir.schemas.OpSignature):
        self._signature = value

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        """Implements an eager-mode execution of an onnxscript function."""
        from onnxscript._internal import evaluator  # pylint: disable=import-outside-toplevel

        return evaluator.default().eval_function(self, args, kwargs)  # type: ignore[arg-type, return-value]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.function!r})"

    def to_function_proto(self) -> onnx.FunctionProto:
        """Converts the function into :class:`onnx.FunctionProto`."""
        return self.function_ir.to_function_proto()

    def graph(self) -> ir.Graph:
        """Returns the IR graph representation of this function.

        Returns:
            The :class:`ir.Graph` representing the computation graph of this function.
            NOTE: This is not a copy, and should not be modified by the caller.
        """
        return self.function_ir.graph

    def to_model_proto(self, **kwargs):
        """Converts the function into :class:`onnx.ModelProto`."""
        if self.function_ir.attrs and any(
            attr.value is None for attr in self.function_ir.attrs
        ):
            raise ValueError(
                "A function with required attributes cannot be exported as a model."
            )
        # Note: The function must also have monomorphic type annotation for inputs/outputs
        # to be converted into a valid model. Otherwise, we can still produce an ONNX
        # model, but it will not pass the ONNX model checker. We do not report an error
        # at this stage.

        # Merge kwargs specified in script-decorator with those specified in this call.
        merged_kw_args = {**self.kwargs, **kwargs}
        return self._to_model_proto(**merged_kw_args)

    def _to_model_proto(
        self,
        io_types: Optional[ONNXType] = None,
        input_types: Optional[Sequence[ONNXType]] = None,
        output_types: Optional[Sequence[ONNXType]] = None,
        value_infos: dict[str, ONNXType] | None = None,
        opset_version: int | None = None,
        **kwargs,
    ) -> onnx.ModelProto:
        """Converts this instance into a `onnx.ModelProto`.

        Args:
            io_types: When specified, all the inputs/outputs of the model
                are set to be of this type.
            input_types: When specified, all the inputs of the model
                are set to be of the corresponding type in this list.
            output_types: When specified, all the outputs of the model
                are set to be of the corresponding type in this list.
            value_infos: A dictionary mapping intermediate variable names to ONNX types.
                Used to set value_info for intermediate variables.
            opset_version: The standard opset version to use for the model if it
                cannot be inferred. Otherwise defaults to the current opset version.
            kwargs: Additional parameters given to function :func:`onnx.helper.make_model`.

        Returns:
            An instance of :class:`onnx.ModelProto`.
        """
        # Identify functions to include in the model
        sub_functions = self.function_ir.get_called_functions()
        ir_functions: list[ir.Function] = [func.function_ir for func in sub_functions.values()]

        # Duplicate the graph to create the model
        main_graph = self.function_ir.graph.clone()
        # Determine opset imports
        opset_imports = main_graph.opset_imports.copy()

        for func in ir_functions:
            domain = func.domain
            if domain is not None and domain not in opset_imports:
                opset_imports[domain] = func.meta.get("opset_version", 1)

            if "" not in opset_imports and "" in func.opset_imports:
                opset_imports[""] = func.opset_imports[""]

        if "" not in opset_imports:
            # No operator is using the standard opset.
            # Use the specified version if provided or the default value.
            opset_imports[""] = (
                opset_version if opset_version is not None else onnx.defs.onnx_opset_version()
            )

        # Determine ir_version
        if "ir_version" in kwargs:
            ir_version = kwargs.pop("ir_version")
        else:
            ir_version = select_ir_version(opset_imports[""])

        # Create the model
        model = ir.Model(main_graph, ir_version=ir_version)
        for func in ir_functions:
            model.functions[func.identifier()] = func

        model_proto = ir.to_proto(model)

        # Update opset imports
        del model_proto.opset_import[:]
        model_proto.opset_import.extend(
            [
                onnx.OperatorSetIdProto(domain=domain, version=version)
                for domain, version in opset_imports.items()
            ]
        )

        # Set additional type information if provided
        graph = model_proto.graph

        if value_infos:
            graph.value_info.extend(
                [
                    onnx.helper.make_value_info(name, type.to_type_proto())
                    for name, type in value_infos.items()
                ]
            )

        if io_types is not None:
            for input in graph.input:
                if not input.HasField("type"):
                    input.type.CopyFrom(io_types.to_type_proto())
            for output in graph.output:
                if not output.HasField("type"):
                    output.type.CopyFrom(io_types.to_type_proto())
        if input_types is not None:
            for input, type in zip(graph.input, input_types):
                input.type.CopyFrom(type.to_type_proto())
        if output_types is not None:
            for output, type in zip(graph.output, output_types):
                output.type.CopyFrom(type.to_type_proto())

        for k, v in kwargs.items():
            setattr(model_proto, k, v)

        return model_proto


class TracedOnnxFunction(Op):
    """TracedOnnxFunction.

    Attributes:
        name: Name of the op. E.g. "aten::add".
        func: Function.
    """

    def __init__(self, opset: Opset, func: Callable):
        super().__init__(opset, func.__name__)
        self.func = func
        self._signature = _schemas.op_signature_from_function(
            self.func, domain="_traced", name=self.name
        )

        # Allow the object to be inspected as a function
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.func!r})"

    @property
    def function_ir(self) -> irbuilder.IRFunction:
        """Return the function_ir.

        This function IR contains only the signature of the function.
        """
        src, func_ast = ast_utils.get_src_and_ast(self.func)
        module = inspect.getmodule(self.func)
        closure = inspect.getclosurevars(self.func)
        global_names = module.__dict__.copy()
        global_names.update(closure.nonlocals)
        converter = converter_module.Converter(
            opset=self._opset,
            global_names=global_names,
            source=src,
        )

        return converter.translate_function_signature(func_ast)

    @property
    def op_signature(self) -> Optional[ir.schemas.OpSignature]:
        """Returns the signature of this op."""
        return self._signature

    @op_signature.setter
    def op_signature(self, value: ir.schemas.OpSignature):
        self._signature = value


class SymbolValue:
    """Represents script-time value information about named variables used in a script.

    At translation-time, the (local) variables of a script, including its parameters,
    are bound to a SymbolValue.

    SymbolValues fall into the following categories:

    AttrRef: Function parameters of attribute-kind, also mapped to ONNX attributes

    Dynamic: values computed at runtime (of tensor type, for now) mapped to NodeArgs.
    Dynamic values include input-parameters of the script, as well intermediate
    values computed in the script.

    For example, consider the following script definition:
    ::

        @script()
        def ThresholdedRelu(X, alpha: float):
            zero = op.CastLike(0, X)
            return op.Where(X > alpha, X, zero)

    Here, `X` has a Dynamic value, `alpha` has an AttrRef value, and `zero`
    has a Dynamic value.

    Scripts may also contain references to global variables, but the translator
    does not associate a SymbolValue with them. The python value of global variables
    is used directly in the translation, and such global variables are intended
    to be used for limited purposes, namely:
    * To identify an opset
    * To represent constant-values, translated into ONNX constants.
    """

    def __init__(self, value: Any, info: sourceinfo.SourceInfo) -> None:
        """
        Initializes SymbolValue.

        Arguments:
            value: The value bound to a python variable in a script.
            info: source-location information for error-messages/debugging
        """
        if not isinstance(info, sourceinfo.SourceInfo):
            raise TypeError(f"info must be of type sourceinfo.SourceInfo not {type(info)!r}.")
        self.value = value
        self.info = info


class AttrRef(SymbolValue):
    def __init__(self, attr: ir.Attr, as_bool: bool, info: sourceinfo.SourceInfo) -> None:
        """Initializes AttrRef.

        Arguments:
            attr: An ir.Attr representing the attribute-parameter
            as_bool: Whether the attribute is to be interpreted as a bool type (represented as int in ONNX)
            info: for debugging use.
        """
        super().__init__(attr, info)
        self.as_bool = as_bool



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
"""torch.ops.aten operators under the `fft` module.

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

from typing import Optional, Sequence

from onnxscript import INT64
from onnxscript.function_libs.torch_lib.registration import torch_op
from onnxscript.function_libs.torch_lib.tensor_typing import TFloat
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import TensorType


def _fftn_onnx_normalization(
    self: TFloat,
    normalization: int,
    signal_size: INT64,
    inverse: bool = False,
) -> TFloat:
    """Normalize in forward or backward direction."""
    # Norm values defined in https://github.com/pytorch/pytorch/blob/758d78790164bfb041555daed380de96e06f78a3/aten/src/ATen/native/SpectralOps.cpp#L117-L131
    # Norm modes: https://github.com/pytorch/pytorch/blob/758d78790164bfb041555daed380de96e06f78a3/aten/src/ATen/native/SpectralOpsUtils.h#L15-L19
    # Modes:
    # 0: no normalization (backward)
    # 1: "ortho" - divide by 1/sqrt(signal_size) (ortho)
    # 2: divide by signal_size (forward)
    signal_size = op.CastLike(signal_size, self)
    if not inverse:
        # Forward normalization
        if normalization == 1:
            self = op.Div(self, op.Sqrt(signal_size))
        elif normalization == 2:
            self = op.Div(self, signal_size)
    else:
        # Backward normalization, accounting for op.DFT already dividing by signal_size
        if normalization == 0:
            self = op.Mul(self, signal_size)
        elif normalization == 1:
            self = op.Mul(self, op.Sqrt(signal_size))
    return self


@torch_op("aten::_fft_c2c", trace_only=True, complex=True)
def aten__fft_c2c(
    self: TFloat, dim: Sequence[int], normalization: int, forward: bool
) -> TFloat:
    """_fft_c2c(Tensor self, SymInt[] dim, int normalization, bool forward) -> Tensor

    Standard complex to complex FFT (forward or backward).
    """

    # NOTE: SymInt dim is not supported because DFT-17 needs a static axis

    # ONNX DFT input assumes the last dimension is the complex dimension.

    unsqueeze_first_dim = 0 in dim
    # 1. Add a new dimension for the end and batch dimension, if needed
    # 2. ONNX DFT input assumes the last dimension is the complex dimension.
    #       If needed, add 1 to account for the batch dimension.

    if unsqueeze_first_dim:
        transformed = op.Unsqueeze(self, axes=[0])
        dim = [d + 1 for d in dim]
    else:
        transformed = self

    for dimension in reversed(dim):
        transformed = op.DFT(transformed, axis=dimension, inverse=not forward, onesided=False)
        transformed = _fftn_onnx_normalization(
            transformed,
            normalization,
            op.Shape(transformed, start=dimension, end=dimension + 1),
            not forward,
        )

    if unsqueeze_first_dim:
        transformed = op.Squeeze(transformed, axes=[0])

    return transformed


@torch_op("aten::_fft_c2r", trace_only=True, complex=True)
def aten__fft_c2r(
    self: TFloat,
    dim: Sequence[int],
    normalization: int,
    last_dim_size: INT64,
) -> TFloat:
    """_fft_c2r(Tensor self, int[] dim, int normalization, SymInt last_dim_size) -> Tensor

    Complex to real inverse FFT. Assumes that input tensor is output of previous FFT operation.
    """

    dimension = dim[0]
    unsqueeze_first_dim = dimension == 0
    # 1. Add a new dimension for batch dimension, if needed
    # 2. ONNX DFT input assumes the last dimension is the complex dimension.
    #       If needed, add 1 to account for the batch dimension.

    if unsqueeze_first_dim:
        transformed = op.Unsqueeze(self, axes=[0])
    else:
        transformed = self

    for idx, dimension in enumerate(dim):
        # Adjust dimension if we unsqueezed at the beginning
        dimension += unsqueeze_first_dim

        if idx < len(dim) - 1:
            transformed = op.DFT(transformed, axis=dimension, inverse=True, onesided=False)
        else:
            # last operation is one-sided, transform to real
            # Torch truncates/pads on the last dimension only. Typically, the only valid values that can be passed
            # into PyTorch are n or n//2+1, where n is self.shape[dim[-1]], but this is not always the case, so we
            # place no such restriction on the ONNX side.
            transformed = op.DFT(
                transformed,
                dft_length=last_dim_size,
                axis=dimension,
                inverse=True,
                onesided=True,
            )

        transformed = _fftn_onnx_normalization(
            transformed,
            normalization,
            op.Shape(transformed, start=dimension, end=dimension + 1),
            inverse=True,
        )

    if unsqueeze_first_dim:
        transformed = op.Squeeze(transformed, axes=[0])

    # Remove the imaginary part
    transformed = op.Slice(transformed, [0], [1], [-1])
    transformed = op.Squeeze(transformed, axes=[-1])

    return transformed


@torch_op("aten::_fft_r2c", trace_only=True)
def aten__fft_r2c(
    self: TFloat, dim: Sequence[int], normalization: int, onesided: bool
) -> TFloat:
    """_fft_r2c(Tensor self, int[] dim, int normalization, bool onesided) -> Tensor

    Real to complex forward FFT.
    """

    # No need to fill the imaginary part because ONNX DFT accepts real inputs
    # https://onnx.ai/onnx/operators/onnx__DFT.html#inputs

    unsqueeze_first_dim = 0 in dim
    # 1. Add a new dimension for the end and batch dimension, if needed
    # 2. ONNX DFT input assumes the last dimension is the complex dimension.
    #       If needed, add 1 to account for the batch dimension.

    if unsqueeze_first_dim:
        transformed = op.Unsqueeze(self, axes=[0, -1])
        dim = [d + 1 for d in dim]
    else:
        transformed = op.Unsqueeze(self, axes=[-1])

    for idx, dimension in enumerate(reversed(dim)):
        transformed = _fftn_onnx_normalization(
            transformed,
            normalization,
            op.Shape(transformed, start=dimension, end=dimension + 1),
            inverse=False,
        )
        if idx > 0:
            transformed = op.DFT(transformed, axis=dimension, inverse=False, onesided=False)
        else:
            # Torch computes one-sided FFT on the last dimension only.
            transformed = op.DFT(transformed, axis=dimension, inverse=False, onesided=onesided)

    if unsqueeze_first_dim:
        transformed = op.Squeeze(transformed, axes=[0])

    return transformed


def aten_fft_fft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_fft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_fft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_fft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_fftfreq(n: int, d: float = 1.0) -> TensorType:
    """fft_fftfreq(int n, float d=1.0, *, ScalarType? dtype=None, Layout? layout=None, Device? device=None, bool? pin_memory=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_fftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_fftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_fftshift(self: TensorType, dim: Optional[int] = None) -> TensorType:
    """fft_fftshift(Tensor self, int[1]? dim=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_hfft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_hfft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_hfft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_hfft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_hfftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_hfftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ifft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_ifft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ifft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_ifft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ifftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_ifftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ifftshift(self: TensorType, dim: Optional[int] = None) -> TensorType:
    """fft_ifftshift(Tensor self, int[1]? dim=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ihfft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_ihfft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ihfft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_ihfft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_ihfftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_ihfftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_irfft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_irfft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_irfft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_irfft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_irfftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_irfftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_rfft(
    self: TensorType, n: Optional[int] = None, dim: int = -1, norm: Optional[str] = None
) -> TensorType:
    """fft_rfft(Tensor self, int? n=None, int dim=-1, str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_rfft2(
    self: TensorType,
    s: Optional[int] = None,
    dim: Sequence[int] = (-2, -1),
    norm: Optional[str] = None,
) -> TensorType:
    """fft_rfft2(Tensor self, int[1]? s=None, int[1] dim=[-2,-1], str? norm=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_rfftfreq(n: int, d: float = 1.0) -> TensorType:
    """fft_rfftfreq(int n, float d=1.0, *, ScalarType? dtype=None, Layout? layout=None, Device? device=None, bool? pin_memory=None) -> Tensor"""

    raise NotImplementedError()


def aten_fft_rfftn(
    self: TensorType,
    s: Optional[int] = None,
    dim: Optional[int] = None,
    norm: Optional[str] = None,
) -> TensorType:
    """fft_rfftn(Tensor self, int[1]? s=None, int[1]? dim=None, str? norm=None) -> Tensor"""

    raise NotImplementedError()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import Any, Iterator

import onnx_ir as ir

from onnxscript._internal import builder as _builder
from onnxscript.nn._parameter import Parameter


class Module:
    """Base class for all onnxscript modules, mirroring PyTorch's nn.Module.

    Subclasses define ``forward()`` to build ONNX subgraphs. Child modules
    and parameters are registered automatically via ``__setattr__``.
    Because ``Parameter`` subclasses ``ir.Value``, parameters like
    ``self.weight`` can be passed directly to ONNX ops.

    Example::

        class Linear(onnxscript.nn.Module):
            def __init__(self, in_features, out_features, bias=True, name=None):
                super().__init__(name)
                self.weight = Parameter([out_features, in_features], name="weight")
                if bias:
                    self.bias = Parameter([out_features], name="bias")
                else:
                    self.bias = None

            def forward(self, op, x):
                w_t = op.Transpose(self.weight, perm=[1, 0])
                result = op.MatMul(x, w_t)
                if self.bias is not None:
                    result = op.Add(result, self.bias)
                return result
    """

    def __init__(self, name: str | None = None) -> None:
        # Use object.__setattr__ to avoid triggering our __setattr__ override
        # before _parameters and _modules dicts exist.
        object.__setattr__(self, "_name", name)
        object.__setattr__(self, "_parameters", {})
        object.__setattr__(self, "_modules", {})

    @property
    def name(self) -> str | None:
        return self._name

    def _set_name(self, name: str) -> None:
        """Set the module name. Subclasses may override to propagate names to children."""
        object.__setattr__(self, "_name", name)

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, Parameter):
            # Auto-register parameters; set default name from attribute name.
            if value.name is None:
                value.name = name
            self._parameters[name] = value
            # Also store on the instance so getattr works outside forward()
            object.__setattr__(self, name, value)
        elif isinstance(value, Module):
            # Auto-register child modules; inherit attribute name if unnamed.
            if value._name is None:
                value._set_name(name)
            self._modules[name] = value
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def __call__(self, op: _builder.OpBuilder, *args: Any, **kwargs: Any) -> Any:
        builder = op.builder
        module_name = self._name or ""
        class_name = type(self).__qualname__
        builder.push_module(module_name, class_name)
        try:
            # Realize parameters: qualify names and register as graph initializers.
            for param in self._parameters.values():
                param._realize(builder)  # pylint: disable=protected-access

            result = self.forward(op, *args, **kwargs)
        finally:
            builder.pop_module()
        return result

    def forward(self, op: _builder.OpBuilder, *args: Any, **kwargs: Any) -> Any:
        """Define the computation performed by this module.

        Must be overridden by subclasses. Receives an ``OpBuilder`` as the
        first argument so that ONNX ops can be called as ``op.MatMul(x, w)``.
        """
        raise NotImplementedError(f"{type(self).__name__} must implement forward()")

    # ------------------------------------------------------------------
    # Iterators
    # ------------------------------------------------------------------

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        """Return an iterator over module parameters."""
        yield from self._parameters.values()
        if recurse:
            for module in self._modules.values():
                yield from module.parameters(recurse=True)

    def named_parameters(
        self, prefix: str = "", recurse: bool = True
    ) -> Iterator[tuple[str, Parameter]]:
        """Return an iterator over module parameters, yielding (name, Parameter) pairs."""
        for name, param in self._parameters.items():
            full_name = f"{prefix}.{name}" if prefix else name
            yield full_name, param
        if recurse:
            for mod_name, module in self._modules.items():
                sub_prefix = f"{prefix}.{mod_name}" if prefix else mod_name
                yield from module.named_parameters(prefix=sub_prefix, recurse=True)

    def children(self) -> Iterator[Module]:
        """Return an iterator over immediate child modules."""
        yield from self._modules.values()

    def named_children(self) -> Iterator[tuple[str, Module]]:
        """Return an iterator over immediate child modules, yielding (name, Module) pairs."""
        yield from self._modules.items()

    def modules(self) -> Iterator[Module]:
        """Return an iterator over all modules in the tree (including self)."""
        yield self
        for module in self._modules.values():
            yield from module.modules()

    def named_modules(self, prefix: str = "") -> Iterator[tuple[str, Module]]:
        """Return an iterator over all modules, yielding (name, Module) pairs."""
        yield prefix, self
        for name, module in self._modules.items():
            sub_prefix = f"{prefix}.{name}" if prefix else name
            yield from module.named_modules(prefix=sub_prefix)

    # ------------------------------------------------------------------
    # State dict
    # ------------------------------------------------------------------

    def state_dict(self, prefix: str = "") -> dict[str, ir.TensorProtocol | None]:
        """Return a dictionary mapping parameter names to their tensor data.

        Mirrors ``torch.nn.Module.state_dict()``. Keys use dot-separated
        hierarchical names (e.g. ``"layer1.weight"``). Values are the
        ``const_value`` of each parameter (``None`` if uninitialized).
        """
        result: dict[str, ir.TensorProtocol | None] = {}
        for name, param in self._parameters.items():
            full_name = f"{prefix}.{name}" if prefix else name
            result[full_name] = param.const_value
        for mod_name, module in self._modules.items():
            sub_prefix = f"{prefix}.{mod_name}" if prefix else mod_name
            result.update(module.state_dict(prefix=sub_prefix))
        return result

    def load_state_dict(
        self,
        state_dict: dict[str, ir.TensorProtocol],
        strict: bool = True,
    ) -> None:
        """Load parameter data from a state dictionary.

        Mirrors ``torch.nn.Module.load_state_dict()``. Sets ``const_value``
        on each matching parameter.

        Args:
            state_dict: Mapping of parameter names to tensor data.
            strict: If ``True`` (default), raises ``KeyError`` for missing
                keys and ``ValueError`` for unexpected keys.
        """
        self._load_state_dict_recursive(state_dict, prefix="", strict=strict)

    def _load_state_dict_recursive(
        self,
        state_dict: dict[str, ir.TensorProtocol],
        prefix: str,
        strict: bool,
    ) -> set[str]:
        """Recursively load state and return the set of consumed keys."""
        consumed: set[str] = set()
        for name, param in self._parameters.items():
            full_name = f"{prefix}.{name}" if prefix else name
            if full_name in state_dict:
                param.const_value = state_dict[full_name]
                consumed.add(full_name)
            elif strict:
                raise KeyError(f"Missing key in state_dict: {full_name!r}")
        for mod_name, module in self._modules.items():
            sub_prefix = f"{prefix}.{mod_name}" if prefix else mod_name
            consumed |= module._load_state_dict_recursive(  # pylint: disable=protected-access
                state_dict, prefix=sub_prefix, strict=strict
            )
        if strict and prefix == "":
            unexpected = set(state_dict.keys()) - consumed
            if unexpected:
                raise ValueError(f"Unexpected keys in state_dict: {unexpected}")
        return consumed

    def __repr__(self) -> str:
        lines = [f"{type(self).__name__}("]
        for name, module in self._modules.items():
            mod_repr = repr(module).replace("\n", "\n  ")
            lines.append(f"  ({name}): {mod_repr}")
        for name, param in self._parameters.items():
            lines.append(f"  ({name}): {param!r}")
        lines.append(")")
        return "\n".join(lines)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Test for MatchContext functionality."""

import unittest

import onnx.parser

from onnxscript import ir
from onnxscript.rewriter import pattern


class MatchContextTest(unittest.TestCase):
    def test_context_usage_in_condition_function(self):
        """Test that MatchContext can be meaningfully used in condition functions."""

        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                c1 = Constant<value_float = 1.0>()
                t1 = Div(c1, x)
                z = Mul(t1, y)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)

        def condition_using_context(context, x, y):
            # Use context to check properties of the match
            self.assertIs(context.model, model)
            self.assertIs(context.graph_or_function, model.graph)
            self.assertIs(context.root, model.graph[2])

            # Verify that we can inspect the matched nodes
            self.assertEqual(len(context.nodes), 2)

            return True  # Allow the rewrite

        def reciprocal_mul_pattern(op, x, y):
            return (1 / x) * y

        def replacement(op, x, y):
            return op.Div(y, x)

        rule = pattern.RewriteRule(
            reciprocal_mul_pattern, replacement, condition_function=condition_using_context
        )

        count = rule.apply_to_model(model)
        self.assertEqual(count, 1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""A utility function to replace custom operations in a model with their expansions"""

from typing import Sequence

import onnx
import onnx_ir as ir
import onnx_ir.passes.common as common_passes


def replace_functions_inplace(irmodel: ir.Model, irfunctions: Sequence[ir.Function]) -> None:
    """A utility function to replace custom operations in a model with their expansions:

    The model is updated in-place.

    Args:
        irmodel: An ONNX model possibly containing calls to custom operations.
        irfunctions: A sequence of functions defining the expansions for the custom operations.


    """
    model_functions = irmodel.functions
    if len(model_functions) != 0:
        # Since we use inlining, check that there are no model-local functions.
        raise ValueError("Input model cannot have model-local functions.")
    for func in irfunctions:
        model_functions[func.identifier()] = func

    # TODO (rama): Ideally, we should provide users more control over renaming strategy for inlined values.
    common_passes.InlinePass()(irmodel)
    common_passes.RemoveUnusedOpsetsPass()(irmodel)


def replace_functions(
    model: onnx.ModelProto, functions: Sequence[onnx.FunctionProto]
) -> onnx.ModelProto:
    """A utility function to replace custom operations in a model with their expansions:
    Args:
        model: An ONNX ModelProto possibly containing calls to custom operations.
        functions: A sequence of FunctionProto defining the expansions for the custom operations.

    Returns:
        An updated ModelProto with custom operations replaced by their expansions.
    """
    irmodel = ir.from_proto(model)
    irfunctions = [ir.from_proto(func) for func in functions]
    replace_functions_inplace(irmodel, irfunctions)
    return ir.to_proto(irmodel)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import math

from onnxscript.rewriter import _fusion_utils, pattern


# Pattern to match against
def erf_gelu_pattern_1(op, x):
    # erf_gelu(x) = 0.5 * x * (1 + erf(x / sqrt(2)))
    # half = pattern.Constant(0.5)
    # sqrt2 = pattern.Constant(1.4142)
    # x_div_sqrt2 = op.Div(x, sqrt2)
    # erf = op.Erf(x_div_sqrt2)
    # one = pattern.Constant(1.0)
    # one_plus_erf = op.Add(erf, one)
    # x_mul_one_plus_erf = op.Mul(x, one_plus_erf)
    # return op.Mul(half, x_mul_one_plus_erf)
    return 0.5 * (x * (op.Erf(x / math.sqrt(2)) + 1.0))


def erf_gelu_pattern_2(op, x):
    return x * (0.5 * (op.Erf(x / math.sqrt(2)) + 1.0))


# Replacement
def gelu(op, x):
    return op.Gelu(x, _domain="com.microsoft")


rule1 = pattern.RewriteRule(erf_gelu_pattern_1, gelu)
rule2 = pattern.RewriteRule(erf_gelu_pattern_2, gelu)

rules = pattern.RewriteRuleSet([rule1, rule2])

fuse_erfgelu = _fusion_utils.apply_fusion_rules(rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Any

import onnx


def normalize_domain(d: str) -> str:
    return "" if d == "ai.onnx" else d


def is_onnx_domain(d: str) -> bool:
    return normalize_domain(d) == ""


def is_onnx_op(node: onnx.NodeProto, op_type: str) -> bool:
    return is_onnx_domain(node.domain) and node.op_type == op_type


def is_control_flow_op(node: onnx.NodeProto) -> bool:
    return any(attr.HasField("g") or len(attr.graphs) > 0 for attr in node.attribute)


def get_node_attr_value(node: onnx.NodeProto, attr_name: str, default: Any) -> Any:
    matching = [x for x in node.attribute if x.name == attr_name]
    if len(matching) > 1:
        raise ValueError(f"Node has multiple attributes with name {attr_name}")
    if len(matching) < 1:
        return default
    return onnx.helper.get_attribute_value(matching[0])


def get_initializer_type(initializer: onnx.TensorProto) -> onnx.TypeProto:
    type = onnx.TypeProto()
    type.tensor_type.elem_type = initializer.data_type
    dims = type.tensor_type.shape.dim
    for dim in initializer.dims:
        dims.add().dim_value = dim
    return type


def get_constant_node_value(node: onnx.NodeProto, name: str) -> onnx.TensorProto | None:
    if (
        node.op_type != "Constant"
        or node.domain not in {"", "ai.onnx"}
        or len(node.attribute) != 1
    ):
        return None
    attr = node.attribute[0]
    if attr.ref_attr_name:
        return None
    attr_name = attr.name
    value = onnx.helper.get_attribute_value(attr)

    if isinstance(value, onnx.TensorProto):
        # Two names exist in this case: we use tensorproto as is (with original name)
        return value
    shape: list[int]
    if attr_name == "value_int":
        dtype = onnx.TensorProto.INT64
        shape = []
        value = [value]
    elif attr_name == "value_float":
        dtype = onnx.TensorProto.FLOAT
        shape = []
        value = [value]
    elif attr_name == "value_string":
        dtype = onnx.TensorProto.STRING
        shape = []
        value = [value]
    elif attr_name == "value_ints":
        dtype = onnx.TensorProto.INT64
        shape = [len(value)]
    elif attr_name == "value_floats":
        dtype = onnx.TensorProto.FLOAT
        shape = [len(value)]
    elif attr_name == "value_strings":
        dtype = onnx.TensorProto.STRING
        shape = [len(value)]
    else:
        return None  # sparse tensors not handled
    return onnx.helper.make_tensor(name, dtype, shape, value)
